package Bind;

# Package to manipulate bind zone files via a DBM database.
# Matthew D. Stock (stock@perdix.acsu.buffalo.edu)
# University at Buffalo
# 2/19/97

use AnyDBM_File;
use POSIX;
use Carp;

sub new {
    my($class) = shift;
    my($self) = {};

    $self->{'version'} = "1.0";

    bless $self, $class;
    $self->load_zones(shift);
    $self->load_config(shift);
    if ($self->tie_zones()) {
	return $self;
    } else {
	return;
    }
}

sub debug {
    my $self = shift;
    $self->{'debug'}=1;
    $self->print_debug("Debug mode enabled");
}

sub print_debug {
    my $self = shift;
    return if !$self->{'debug'};

    printf @_;
    print "<BR>\n";
}

sub load_config {
    my($self) = shift;
    my($file) = shift;
    my($token,@tmp);

    $self->{'dbm_dir'} = "named/dbm";
    $self->{'tmp_dir'} = "/var/tmp";
    $self->{'log_file'} = "/var/tmp/dns.log";
    $self->{'hostmaster'} = 'stock@acsu.buffalo.edu';
    $self->{'RRlist'} = ['soa', 'ns', 'a', 'mx', 'ptr', 'hinfo', 'rp', 'txt', 
			 'cname'];
    $self->{'killcache'} = 1;

    open(CONFIG,$file) || return 0;
    while (<CONFIG>) {
	chop;
	@_ = split;
	$token = shift;
	if ($token eq "dbm_dir") {
	    $self->{'dbm_dir'} = shift;
	    next;
	}
	if ($token eq "hostmaster") {
	    $self->{'hostmaster'} = shift;
	    next;
	}
	if ($token eq "RR") {
	    @tmp = split(',',shift);
	    $self->{'RRlist'} = \@tmp;
	}
	if ($token eq "user") {
	    @tmp = split(',',shift);
	    $self->{'user_list'} = \@tmp;
	    next;
	}
	if ($token eq "admin") {
	    @tmp = split(',',shift);
	    $self->{'admin_list'} = \@tmp;
	    next;
	}
	if ($token eq "tmp_dir") {
	    $self->{'tmp_dir'} = shift;
	    next;
	}
	if ($token eq "log_file") {
	    $self->{'log_file'} = shift;
	    next;
	}
	if ($token eq "backup_hist") {
	    $self->{'backup_hist'} = shift;
	    next;
	}
	if ($token eq "backup_dir") {
	    $self->{'backup_dir'} = shift;
	    next;
	}
    }
    close(CONFIG);
    return 1;
}

sub tie_zones {
    my($self) = shift;
    my($failed) = 0;

    foreach (@{$self->{'RRlist'}}) {
	if (!tie(%{$self->{"in_$_"}},AnyDBM_File,$self->{'dbm_dir'}.'/in_'.$_,
		 &POSIX::O_RDWR, 0644)) {
	    $failed = 1;
	}
    }

    if (!tie(%{$self->{'zone'}},AnyDBM_File,$self->{'dbm_dir'}.'/zone',
	     &POSIX::O_RDWR, 0644)) {
	$failed = 1;
    }
    if (!tie(%{$self->{'dirty'}},AnyDBM_File,$self->{'dbm_dir'}.'/dirty',
	     &POSIX::O_RDWR, 0644)) {
	$failed = 1;
    }
    if (!tie(%{$self->{'hinfo_set'}},
	     AnyDBM_File,$self->{'dbm_dir'}.'/hinfo_set',
	     &POSIX::O_RDWR, 0644)) {
	$failed = 1;
    }

    if ($failed) {
	return $self->zone_to_dbm();
    } else {
	return 1;
    }
}

# Test & set a lock for zones.
sub set_lock {
    my($self) = shift;
    my($zone) = shift;
    my($tmpfile) = $self->{'dbm_dir'}."/$zone.".time().'.tmp';
    my($lockfile) = $self->{'dbm_dir'}."/$zone.lock";
    my($happy);

    open(FOO,">$tmpfile") || return 0;
    # Ok, let's try the link, as it's atomic
    $happy = link($tmpfile,$lockfile);
    unlink($tmpfile);
    return $happy
}

# Delete the lockfile
sub unset_lock {
    my($self) = shift;
    my($zone) = shift;
    
    unlink($self->{'dbm_dir'}."/$zone.lock");
}

sub untie_zones {
    my($self) = shift;

    foreach (@{$self->{'RRlist'}}) {
	untie(%{$self->{$_}});
    }
    untie(%{$self->{'zone'}});
    untie(%{$self->{'hinfo_set'}});
    untie(%{$self->{'dirty'}});
}

sub zone_to_dbm {
    my($self) = shift;

    if (!$self->set_lock('dbm')) {
	return 0;
    }

    $self->untie_zones;
    $self->delete_zones;

    foreach (@{$self->{'RRlist'}}) {
	tie(%{$self->{"in_$_"}},AnyDBM_File,$self->{'dbm_dir'}.'/in_'.$_,
	    &POSIX::O_RDWR|&POSIX::O_CREAT, 0644);
    }

    tie(%{$self->{'zone'}},AnyDBM_File,$self->{'dbm_dir'}.'/zone',
	&POSIX::O_RDWR|&POSIX::O_CREAT, 0644) || croak "Can't tie to ".
	    $self->{'dbm_dir'}."zone: $!";
    tie(%{$self->{'dirty'}},AnyDBM_File,$self->{'dbm_dir'}.'/dirty',
	&POSIX::O_RDWR|&POSIX::O_CREAT, 0644) || croak "Can't tie to ".
	    $self->{'dbm_dir'}."dirty: $!";
    tie(%{$self->{'hinfo_set'}},AnyDBM_File,$self->{'dbm_dir'}.'/hinfo_set',
	&POSIX::O_RDWR|&POSIX::O_CREAT, 0644) || croak "Can't tie to ".
	    $self->{'dbm_dir'}."hinfo_set: $!";

    while (($zone, $file) = each %{ $self->{'zones'}}) {
	$self->load_zonefile($file, $zone);
    }
    # We do this as a 2nd pass so out-of-zone data doesn't "dirty" other
    # zones before we've even made any changes.
    while (($zone, $file) = each %{ $self->{'zones'}}) {
	$self->{'dirty'}{$zone} = 0;
    }

    return $self->unset_lock('dbm');
}

sub delete_zones {
    my($self) = shift;
    my($dbm_dir) = $self->{'dbm_dir'};
    
    $self->untie_zones;

    foreach (sort <$dbm_dir/dirty* $dbm_dir/in_* $dbm_dir/zone* $dbm_dir/hinfo_set*>) {
	unlink $_ ;
    }
}

sub load_zonefile {
    my($self) = shift;
    my($zonefile, $origin) = @_;
    my($domain) = $origin;
    my($class,$ttl,$rtype,$tmp,$cont,$i);

    open(ZONE,"$zonefile") || croak "Can't open $zonefile: $!.";

    while (<ZONE>) {
	undef($ttl);
	chop;

	($cont, $_) = $self->strip_comment($_);

	while ($cont) {
	    $_ .= <ZONE>;  chop;
	    ($cont, $_) = $self->strip_comment($_);
	}

	next if ($_ eq "");

	@_ = split;
	$tmp = shift;

	if ($tmp eq '') { next; }
	if ($tmp eq '$ORIGIN') { $origin = shift; next; }
	if ($tmp eq '$INCLUDE') { next; } # We don't do INCLUDES...
	if ($tmp eq '@') {
	    $domain = $origin;
	    $tmp = shift;
	} elsif ($tmp ne 'IN') {
	    $domain = $self->fqdn($tmp,$origin);
	    $tmp = shift;
	}
	if ($tmp =~ /^[0-9]+$/) {
	    $ttl = $tmp;
	    $tmp = shift;
	}
	if ($tmp eq 'IN') {
	    $class = $tmp;
	    $rtype = shift;
	    
	    # Because of some legal (but gross) stuff in the zone files,
	    # Sometimes data is "misplaced" in the parent domain.
	    # We must test to determine what zone the domain actually is in
	    # by looking at the delegations.
	    @tmp = split(/\./,$domain);
	    while (@tmp) {
		my($sub) = join('.',@tmp);
		if ($self->{'seczones'}{$sub}) {
		    $self->{'zone'}{$domain} = $sub;
		    last;
		} elsif ($self->{'zones'}{$sub}) {
		    $self->{'zone'}{$domain} = $sub;
		}
		shift @tmp;
	    }
	    if (!$self->{'zone'}{$domain}) {
		$self->{'zone'}{$domain} = $origin;
	    }
	    $self->add($domain,$origin,$ttl,uc($rtype),join(' ',@_));
	}
    }
}

sub hinfo_parse {
    my($self) = shift;

    return join(' ',@_);
}

sub cname_parse {
    my($self) = shift;
    my($origin) = shift;

    return $self->fqdn(shift,$origin);
}

sub txt_parse {
    my($self) = shift;
    my($part);

    return '"' . join(' ',grep(/[^\"]/,@_)) . '"';
}

sub rp_parse {
    my($self) = shift;
    my($origin) = shift;

    return $self->fqdn(shift,$origin) . " " . $self->fqdn(shift,$origin);
}

sub soa_parse {
    my($self) = shift;
    my($origin,$soa,$contact,@data) = @_;

    $soa = $self->fqdn($soa,$origin);
    $contact = $self->fqdn($contact,$origin);
    
    return "$soa $contact " . join(' ',@data);
}

sub fqdn {
    my($self) = shift;
    my($domain, $origin) = @_;
    
    if (defined($domain) && $domain !~ /\.$/) {
	$domain .= ".$origin";
    }
    return $domain;
}


sub strip_comment {
    my($self) = shift;
    my($line) = shift;
    my($cont) = 0;

    if ($line =~ /([^;]*);/o) { $line = $1; } # Strip comments
    if ($line !~ /\(/o &&
	$line !~ /\"/o) { return (0, $line); } # Shortcut if we can...
    if ($line !~ /\([^\)]*\)/o) { $cont = 1; } # Mismatched parens?
    if ($line !~ /(\"[^\"]*\")*/o) { $cont = 1; } # Mismatched quotes?

    return ($cont,$line);
}

sub load_zones {
    my($self) = shift;
    my($directory,$token);

    if (open(BOOT,shift)) {
	while (<BOOT>) {
	    chop; s/;.*//; @_ = split;
	    $token = shift;
	    if ($token eq "directory") {
		$directory = shift;
		next;
	    }
	    if ($token eq "primary") {
		($zone, $file) = @_;
		$self->{'zones'}{"$zone."} =
		    "$directory/$file";
		next;
	    }
	    if ($token eq "secondary") {
		($zone, $ip, $file) = @_;
		$self->{'seczones'}{"$zone."} =
		    [ $ip, "$directory/$file" ];
		next;
	    }
	}
	close(BOOT);
	return 1;
    }
    return 0;
}

# To confuse people we have 3 similarly-named functions that do different
# things.  listzone give the user an array of the hosts contained within a
# zone.
sub listzone {
    my($self) = shift;
    my($zone) = shift;
    my(@list);

    foreach (keys %{$self->{'zone'}}) {
	if ($self->{'zone'}{$_} eq $zone) {
	    push(@list, $_);
	}
    }
    return sort zonesort @list;
}

# lists records matching certain criterion.
sub list {
    my($self) = shift;
    my($host) = shift;
    my($scope) = lc(shift);
    my(@results,@tmp,$tmp);

    @tmp = split(/\./,$host);
    $i = $#tmp;

    while (@tmp) {
	my($tmp) = join('.',@tmp).'.';
	if ($self->{'seczones'}{$tmp}) {
	    return "ns:".$self->{'seczones'}{$tmp}[0];
	}
	shift @tmp;
    }

    if ($scope eq "all") {
	foreach (@{$self->{'RRlist'}}) {
	    push (@results,@tmp) if (@tmp = $self->list($host,lc($_)));
	}
	return @results;
    }

    if ($scope =~ /^(mx|a|ptr|txt|ns)$/) {
	for ($i=0; $self->{"in_$scope"}{"$host;$i"}; $i++) {
	    push(@results,"$scope:".$self->{"in_$scope"}{"$host;$i"});
	}
	return @results if (@results);
	return;
    }
    
    return ("$scope:".$self->{"in_$scope"}{$host}) 
	if ($self->{"in_$scope"}{$host});
    return;
}

# Return a list of either primary of secondary zones.
sub zonelist {
    my($self) = shift;
    my($type) = shift; # PRIMARY or SECONDARY
    my(@tmp);

    if (uc($type) eq "PRIMARY") {
	foreach (keys %{$self->{'zones'}}) {
	    push(@tmp,$_);
	}	
    }
    if (uc($type) eq "SECONDARY") {
	foreach (keys %{$self->{'seczones'}}) {
	    push(@tmp,$_);
	}	
    }
    return sort zonesort @tmp;
}

# Add a new record to the database.  This may overwrite some kinds of RRs
# like (SOAs) that only can have one, or it appends (A, MX, etc).
# Calls are of the form:
#      add("foo.buffalo.edu.",<undef>,"buffalo.edu.","A","128.205.7.1")
sub add {
    my($self,$domain,$origin,$ttl,$rtype,$contents) = @_;
    my(@tmp,$tmp,$i);

    $self->print_debug("add name=$domain zone=$origin ttl=$ttl rtype=$rtype contents=$contents");


    @_ = split(/\s/,$contents);

    @tmp = split(/\./,$host);
    $i = $#tmp;

    while (@tmp) {
	my($tmp) = join('.',@tmp).'.';
	if ($self->{'seczones'}{$tmp}) {
	    return -1;
	}
	shift @tmp;
    }

    if (!defined($self->{'zone'}{$domain})) {
	$self->{'zone'}{$domain} = $origin;
    }

    if (uc($rtype) eq 'SOA') {
	$self->{'in_soa'}{$domain} = $self->soa_parse($origin,@_);
	$self->{'dirty'}{$origin} = 1;
	return 1;
    }
    if (uc($rtype) eq 'A') {
	my($ip) = shift;
	my($skip) = 0;
	for ($i=0; $self->{'in_a'}{"$domain;$i"}; $i++) {
	    ($tmpttl, $tmpip) = split(/\s/,$self->{'in_a'}{"$domain;$i"});
	    if ($tmpip eq $ip && $tmpttl eq $ip) {
		$skip = 1;
	    }
	}
	
	if (!$skip) {
	    $self->{'in_a'}{"$domain;$i"} = 
		defined($ttl) ? "$ttl $ip" : $ip;
	    $self->{'dirty'}{$self->{'zone'}{$domain}} = 1;
	}
	return 1;
    }
    if (uc($rtype) eq 'NS') {
	my($host) = $self->fqdn(shift,$origin);
	my($skip) = 0;
	
	for ($i=0; $self->{'in_ns'}{"$domain;$i"}; $i++) {
	    if ($self->{'in_ns'}{"$domain;$i"} eq $host) {
		$skip = 1;
	    }
	}
	
	if (!$skip) {
	    $self->{'in_ns'}{"$domain;$i"} = $host;
	    $self->{'dirty'}{$self->{'zone'}{$domain}} = 1;
	}
	return 1;
    }
    if (uc($rtype) eq 'RP') {
	$self->{'in_rp'}{$domain} = $self->rp_parse($origin,@_);
	$self->{'dirty'}{$self->{'zone'}{$domain}} = 1;
	return 1;
    }
    if (uc($rtype) eq 'TXT') {
	for ($i=0; $self->{'in_txt'}{"$domain;$i"}; $i++) {};
	$self->{'in_txt'}{"$domain;$i"} = $self->txt_parse(@_);
	$self->{'dirty'}{$self->{'zone'}{$domain}} = 1;
	return 1;
    }
    if (uc($rtype) eq 'PTR') {
	my($skip) = 0;
	my($host) = $self->fqdn(shift,$origin);

	for ($i=0; $self->{'in_ptr'}{"$domain;$i"}; $i++) {
	    if ($self->{'in_ptr'}{"$domain;$i"} eq $host) {
		$skip = 1;
	    }
	}
	
	if (!$skip) {
	    $self->{'in_ptr'}{"$domain;$i"} = $host;
	    $self->{'dirty'}{$self->{'zone'}{$domain}} = 1;
	}
	return 1;
    }
    if (uc($rtype) eq 'CNAME') {
	$self->{'in_cname'}{$domain} = $self->cname_parse($origin,@_);
	$self->{'dirty'}{$self->{'zone'}{$domain}} = 1;
	return 1;
    }
    if (uc($rtype) eq 'HINFO') {
	$self->{'in_hinfo'}{$domain} = $self->hinfo_parse(@_);
	$self->{'hinfo_set'}{$self->{'in_hinfo'}{$domain}}++;
	$self->{'dirty'}{$self->{'zone'}{$domain}} = 1;
	return 1;
    }
    if (uc($rtype) eq 'MX') {
	my($skip) = 0;
	my($pref) = shift;
	my($host) = $self->fqdn(shift,$origin);
	    
	for ($i=0; $self->{'in_mx'}{"$domain;$i"}; $i++) {
	    if ($self->{'in_mx'}{"$domain;$i"} eq "$pref $host") {
		$skip = 1;
	    }
	}
	
	if (!$skip) {
	    $self->{'in_mx'}{"$domain;$i"} = "$pref $host";
	    $self->{'dirty'}{$self->{'zone'}{$domain}} = 1;
	}
	return 1;
    }
    return 0;
}

sub delete {
    my($self) = shift;
    my($domain) = shift;
    my($scope) = lc(shift);
    my($contents) = shift;
    my(@tmp,$tmp,@results,$j,$i);
    
    @tmp = split(/\./,$host);
    $i = $#tmp;

    while (@tmp) {
	my($tmp) = join('.',@tmp).'.';
	if ($self->{'seczones'}{$tmp}) {
	    return -1;
	}
	shift @tmp;
    }

    if ($scope eq 'all') {
	foreach (@{$self->{'RRlist'}}) {
	    push(@results,@tmp) if (@tmp = $self->delete($domain,$_));
	}
	delete($self->{'zone'}{$domain});
	return @results;
    }

    if ($scope =~ /^(mx|a|ptr|txt|ns)$/) {
	if (!$contents) {
	    for ($i=0; $self->{"in_$scope"}{"$domain;$i"}; $i++) {
		push(@results, "$scope:".$self->{"in_$scope"}{"$domain;$i"});
		delete($self->{"in_$scope"}{"$domain;$i"});
		$self->{'dirty'}{$self->{'zone'}{$domain}} = 1;
	    }
	}
	for ($i=0; $self->{"in_$scope"}{"$domain;$i"}; $i++) {
	    if ($self->{"in_$scope"}{"$domain;$i"} eq $contents) {
		push(@results, "$scope:".$self->{"in_$scope"}{"$domain;$i"});
		for ($j=$i+1; $self->{"in_$scope"}{"$domain;$j"}; $i++, $j++) {
		    $self->{"in_$scope"}{"$domain;$i"} = 
			$self->{"in_$scope"}{"$domain;$j"};
		}
		delete($self->{"in_$scope"}{"$domain;$i"});
		$self->{'dirty'}{$self->{'zone'}{$domain}} = 1;
		last;
	    }
	}
	return @results;
    }

    if (!$contents) {
	$tmp = "$scope:".$self->{"in_$scope"}{"$domain"};
	delete($self->{"in_$scope"}{"$domain"});
	if ($scope eq 'hinfo') {
	    if (--$self->{'hinfo_set'}{$self->{'in_hinfo'}{$domain}} < 1) {
		delete($self->{'hinfo_set'}{$self->{'in_hinfo'}{$domain}});
	    }
	}
	$self->{'dirty'}{$self->{'zone'}{$domain}} = 1;
	return ($tmp);
    }

    if ($self->{"in_$scope"}{$domain} eq $contents) {
	$tmp = "$scope:$contents";
	delete($self->{"in_$scope"}{$domain});
	if ($scope eq 'hinfo') {
	    if (--$self->{'hinfo_set'}{$self->{'in_hinfo'}{$domain}}) {
		delete($self->{'hinfo_set'}{$self->{'in_hinfo'}{$domain}});
	    }
	}
	$self->{'dirty'}{$self->{'zone'}{$domain}} = 1;
	return ($tmp);
    }
    return;
}

# Returns an array of the in-use hinfo settings.  It speeds up processing,
# and also reduces user error doing it this way instead of with a text
# box.  The set keeps a refernce count, and once the last HINFO using
# that string is deleted, it is deleted.
sub list_hinfo {
    my($self) = shift;

    return keys %{$self->{'hinfo_set'}};
}

# Dumps dirty zones in zone file format, and flags the server for a reload.
# We return a list of dumped zones.
sub dump {
    my($self) = shift;
    my(@dirty_list,@soa);

    foreach (keys %{$self->{'zones'}}) {
	if ($self->{'dirty'}{$_}) {
	    push(@dirty_list,$_);
	    @soa = split(/ /,$self->{'in_soa'}{$_});
	    $soa[3] = $self->update_soa($soa[3]);
	    $self->{'in_soa'}{$_} = join(' ',@soa);
	    $self->{'dirty'}{$_} = 
		!$self->save_zonefile($_,$self->{'zones'}{$_});
	}
    }
    # If no dirty_list, then kill DBM cache
    if (! @dirty_list && $self->{'killcache'}) {
	$self->delete_zones;
    }
    return @dirty_list;
}

sub zonesort {
    my(@a) = split(/\./,$a);
    my(@b) = split(/\./,$b);

    while (@a) {
        $aval = pop @a;
        $bval = pop @b;
        if ($aval =~ /^\d+$/ && $bval =~ /^\d+$/) {
            if ($foo = $aval <=> $bval) {
                return $foo;
            }
        } else {
            if ($foo = $aval cmp $bval) {
                return $foo;
            }
        }
    }
    return 0;
}

# Write data for a given zone to a given file in zone file format.
# We make a backup copy of the zone file in the 'backup_dir', and rotate
# after 'backup_hist'.
sub save_zonefile {
    my($self) = shift;
    my($zone) = shift;
    my($filename) = shift;
    my($version) = $self->{'version'};
    my(@records,$head,$name,$ref);

    open(ZONE,"> $filename.tmp") || return 0;

    print ZONE "; Generated by Bind.pm v$version.\n";
    print ZONE "; Author: Matthew D. Stock (stock\@acsu.buffalo.edu)\n";
    print ZONE "; URL: http://perdix.acsu.buffalo.edu/~stock/webdns.html\n";
    print ZONE ";-------------------------------------------------------\n";

    print ZONE "@\t\tIN SOA\t",$self->{'in_soa'}{$zone},"\n";

    # We first throw the zone data into an array so we can sort it.
    foreach $name (@{$self->{'RRlist'}}) {
	# We ignore SOA data because we should only have one, and we printed it
	# up above.
	if ($name eq "soa") {
	    next;
	}
	foreach (keys %{$self->{"in_$name"}}) {
	    my($domain, $i) = split(/;/);
	    if ($self->{'zone'}{$domain} eq $zone) {
		$domain =~ s/\.?$zone$//;
		push(@{$records{$domain}}, 
		     [uc $name, $self->{"in_$name"}{$_}]);
	    }
	}
    }

    # We sort by the hostname, and print the records out.
    # We only print out the hostname once if there are multiple records.
    foreach $name (sort zonesort keys %records) {
	$head = $name;
	$head = "\t" if $head eq "";
	foreach $ref (@{$records{$name}}) {
	    print ZONE "$head\tIN $ref->[0] $ref->[1]\n";
	    $head = "\t";
	}
    }
    close(ZONE);
    return 1;
}

# Given a serial number, return a number greater than it in date format
# (if possible).
sub update_soa {
    my($self) = shift;
    my($oldval) = shift;
    my(@currdate) = (localtime(time))[5,4,3];
    my($today, $inc);
    
    $currdate[0] += 1900; # We want 4 digit years (localtime inc 100's).

    # Create the new serial number
    $today = sprintf "%4d%2d%2d",$currdate[0],$currdate[1]+1,$currdate[2];

    if (length $today eq 10 && $today eq substr($oldval,0,8)) {
	$inc = substr($oldval,8,2);
	$inc++; # We need to do a string increment, not add 1 above.
    } else {
	$inc = '01';
    }
    $today .= $inc;

    if ($today <= $oldval) {
	return $oldval+1;
    } else {
	return $today;
    }
}

return 1;
